﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32.SafeHandles;
using MU;

namespace TOF
{
    public class Cal
    {
        public static string[] DataOutReg = { "Data Invalid", "Distance MSB", "Distance MSB", "Percision MSB", "Percision LSB",
                                                "Magnitude EXP", "Magnitude MSB", "Magnitude LSB", "Phase EXP", "Phase MSB", 
                                                "Phase LSB", "I RAW EXP", "I RAW MSB", "I RAW LSB", "Q RAW EXP", "Q RAW MSB", 
                                                "Q RAW LSB", "Emitter Volt Before", "Emitter Volt After", "AFE Temp", "Ambient ADC", 
                                                "VGA1", "VGA2", "GAIN MSB", "GAIN LSB", "I ADC MSB", "I ADC LSB", "Q ADC MSB", "Q ADC LSB"};

        public static void XtalkCal(TofMain inTofMain, int N)
        // This method measures crosstalk and averages the value N times
        // the average is calculated and written offset registers
        {
            double result = 0;
            double dIrawSum = 0;                // doubles for output data I, Q & Gain sums
            double dQrawSum = 0;                // doubles for output data I, Q & Gain sums
            double dGainSum = 0;                // doubles for output data I, Q & Gain sums
            byte[] reg_3_bytes = new byte[3];
            //byte[] num_8_bytes = new byte[] { 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0 };
            byte[] reg_bytes = new byte[27];
            byte reg0x13 = Comm.I2C_read(0x13); // Save Sample Control settings
            byte reg0x60 = Comm.I2C_read(0x60); // Save Interrupt Control settings
            Comm.I2C_write(0x13, 0x7D);         // Configure sample control
            Comm.I2C_write(0x60, 0x01);         // Enable Data Ready interrupt only

            TofMain.tofSyncDataStruct[] curSyncData = new TofMain.tofSyncDataStruct[1];
            TofMain.tofSyncDataStruct[] DataArray = new TofMain.tofSyncDataStruct[10];
            result = 0;


            int a;
            for (a = 0; a < N; ++a)
            {
                //Comm.GetOutputReg(reg_bytes);
                //inTofMain.SingleShot(TofMain.curDeviceHandle);    // modify to combine next 2 statements
                result = inTofMain.Get501SyncData(curSyncData);
                Comm.GetOutputReg(inTofMain, reg_bytes);
                dIrawSum += convert.Bytes2Double3S(reg_bytes[10], reg_bytes[11], reg_bytes[12]);
                dQrawSum += convert.Bytes2Double3S(reg_bytes[13], reg_bytes[14], reg_bytes[15]);
                dGainSum += convert.Bytes2Double2U(reg_bytes[22], reg_bytes[23]);


                byte[] baResult = new byte[8];
                baResult = BitConverter.GetBytes(2.0);
            }
            dIrawSum = dIrawSum / N;
            dQrawSum = dQrawSum / N;
            dGainSum = dGainSum / N;

            convert.Double2Bytes3S(dIrawSum, ref reg_3_bytes);
            convert.ByteNumWrite(0x24, ref reg_3_bytes, 3); // Write dIrawSum in 3 byte format
            convert.Double2Bytes3S(dQrawSum, ref reg_3_bytes);
            convert.ByteNumWrite(0x27, ref reg_3_bytes, 3); // Write dIrawSum in 3 byte format
            convert.Double2Bytes2U(dGainSum, ref reg_3_bytes);
            convert.ByteNumWrite(0x2A, ref reg_3_bytes, 2); // Write dIrawSum in 3 byte format

            Comm.I2C_write(0x13, reg0x13);      // Restore Sample Control settings
            Comm.I2C_write(0x60, reg0x60);      // Restore Interrupt Control settings
        }

        public static double DistCal(TofMain inTofMain, double ActDist, int N)
        // This method measures phase offset and averages the value N times
        // the average is calculated and written phase offset registers
        {
            double result = 0;
            double dDistSum = 0;                // double for distance sum
            byte[] reg_bytes = new byte[27];    // holds the latest meas outputs


            byte reg0x13 = Comm.I2C_read(0x13); // Save Sample Control settings
            byte reg0x60 = Comm.I2C_read(0x60); // Save Interrupt Control settings

            Comm.I2C_write(0x13, 0x7D);         // Configure sample control
            Comm.I2C_write(0x60, 0x01);         // Enable Data Ready interrupt only

            TofMain.tofSyncDataStruct[] curSyncData = new TofMain.tofSyncDataStruct[1];
            TofMain.tofSyncDataStruct[] DataArray = new TofMain.tofSyncDataStruct[10];

            int a;
            for (a = 0; a < N; ++a)
            {
                result = inTofMain.Get501SyncData(curSyncData);
                Comm.GetOutputReg(inTofMain, reg_bytes);
                dDistSum += (double)((reg_bytes[8] << 8) + reg_bytes[9]);
            }
            dDistSum = dDistSum / N;            // calculate average
            dDistSum = dDistSum - (ActDist /33.3 * 65536);      // subtract actual distance to get phase offset

            Comm.I2C_write(0x2F, (byte)(((int)dDistSum & 0xFF00) >> 8));    // Write phase offset MSB
            Comm.I2C_write(0x30, (byte)((int)dDistSum & 0x00FF));           // Write phase offset LSB

            Comm.I2C_write(0x13, reg0x13);      // Restore Sample Control settings
            Comm.I2C_write(0x60, reg0x60);      // Restore Interrupt Control settings

            dDistSum = dDistSum * 33.3 / 65536; // convert phase to distance

            return dDistSum;
        }

        public static void SS_2_file_N(TofMain inTofMain, string filename, int N)
        // This method measures crosstalk and averages the value N times
        // the average is calculated and written offset registers
        {
            StreamWriter sw = null;  // must assign an initial value
            try
            {
                //string filePath = System.IO.Path.GetFullPath("TestFile.txt"); 
                sw = new StreamWriter(filename, false, Encoding.ASCII);  // open file 
            }
            catch (IOException)
            {
                Console.WriteLine("\nCannot write because the file is open.\n");
                Console.WriteLine("Close file or change the output name.\n");
                Environment.Exit(1);
            }

            Console.WriteLine(DateTime.Today.ToLongDateString() + "   " + DateTime.Now.ToLongTimeString());
            sw.WriteLine(DateTime.Today.ToLongDateString() + "   " + DateTime.Now.ToLongTimeString());


            double result = 0;
            byte[] reg_bytes = new byte[27];
            byte reg0x13 = Comm.I2C_read(0x13); // Save Sample Control settings
            byte reg0x60 = Comm.I2C_read(0x60); // Save Interrupt Control settings
            Comm.I2C_write(0x13, 0x7D);         // Configure sample control
            Comm.I2C_write(0x60, 0x01);         // Enable Data Ready interrupt only

            TofMain.tofSyncDataStruct[] curSyncData = new TofMain.tofSyncDataStruct[1];
            TofMain.tofSyncDataStruct[] DataArray = new TofMain.tofSyncDataStruct[10];

            int a;
            int b;
            for (b = 0; b < reg_bytes.Length; ++b)  // write register names to file
                sw.Write(DataOutReg[b] + ",");
            sw.Write("\r\nTest#,");
            for (b = 0xD0; b < 0xEB; ++b)           // write register numbers to file
                sw.Write(string.Format("{0:x},", b));
            sw.WriteLine();

            for (a = 0; a < N; ++a)
            {
                result = inTofMain.Get501SyncData(curSyncData);
                Comm.GetOutputReg(inTofMain, reg_bytes);
                sw.Write(string.Format("{0:N},", a+1));
                for (b = 0; b < reg_bytes.Length; ++b)
                {
                    sw.Write(reg_bytes[b].ToString() + ",");
                }
                sw.WriteLine();
            }
            sw.Close();
            Comm.I2C_write(0x13, reg0x13);      // Restore Sample Control settings
            Comm.I2C_write(0x60, reg0x60);      // Restore Interrupt Control settings
        }

    }



    public class Comm
    {
        public static void I2C_write(int reg_add, int reg_data)
        // this method writes a single register, one byte address, one byte data
        {
            TofMain.WriteRegToDevice(reg_add, (byte)reg_data);
        }
        public static byte I2C_read(int reg)
        // this method reads a single register, one byte address, one byte data
        {
            byte data = 0;
            TofMain.ReadRegFromDevice(reg);
            return data;
        }
        public static void GetOutputReg(TofMain inMainForm,byte[] data_bytes)
        {
            //TofMain get = new TofMain();    // create instance of TofMain

            inMainForm.Get501RegData(TofMain.curDeviceHandle, 0xD0, 27, data_bytes);

        }
    }

    public class convert
    {
        public static double Bytes2Double3S(byte exp, byte msb, byte lsb)
        // method converts 3 signed bytes in format {Exp,MSB,LSB} to a double 
        {
            double result = 0;
            int iMantissa = 0;                
            bool negative = false;  // flag for negative numbers

            if (msb > 127)
                negative = true;    // negative number

            iMantissa = msb << 8;
            iMantissa |= lsb;
            if (negative)
            {
                iMantissa = ((iMantissa - 1) ^ 0xFFFF);   // convert from 2's complement
                result = -iMantissa * Math.Pow(2, exp);               // combine mantissa and exponent
            }
            else
                result = iMantissa * Math.Pow(2, exp);

            return result;
        }


        public static double Bytes2Double2U(byte msb, byte lsb)
        // method converts bytes in format {MSB,LSB} to a double 
        {
            double result = 0;
            result = (double)(ushort)(((int)(msb) << 8) | (int)(lsb));
            return result;
        }

        public static void Double2Bytes3S(double dNum, ref byte[] baResult)
        // method converts a double to 3 bytes with signed mantissa {Exp,MSB,LSB} 
        {
            double dNumLog = 0;
            byte a;
            bool negative = false;
            byte exp, new_exp = 0;
            int iMantissa = 0;
            double dMantissa = 0;

            if (dNum < 0)       // handle negative numbers
            {
                negative = true;        // set negative flag
                dNum = Math.Abs(dNum);  // convert to positive
            }

            dNumLog = Math.Log(dNum, 2);            // log base 2 of input
            exp = (byte)dNumLog;                    // exponent of the double
            dMantissa = (dNumLog - (double)exp);    // log of mantissa
            dMantissa = Math.Pow(2, dMantissa);     // convert mantissa to double
            new_exp = exp;                          // start new exponent as the original exponent

            // it might seem like 15 shifts is the correct number but it's 14. 
            // Doing 15 shifts into the sign bit making it a negative number.
            for (a = 1; a <= exp && a < 15; ++a)    // convert mantissa to whole number
            {
                dMantissa = dMantissa * 2;          // double the mantissa
                --new_exp;                          // decrement the exponent
            }

            if(negative)
                iMantissa = (int)(-dMantissa);     // take 2's complement, convert to short
            else
                iMantissa = (int)(dMantissa);      // convert to short

            baResult[0] = new_exp;
            baResult[1] = (byte)((iMantissa & 0xFF00) >> 8);
            baResult[2] = (byte)(iMantissa & 0x00FF);
        }

        public static void Double2Bytes2U(double dNum, ref byte[] baResult)
        // method converts a double to 2 bytes {MSB,LSB} 
        {
            int iNum = (int)dNum;
            baResult[0] = (byte)((iNum & 0x0000FF00) >> 8);
            baResult[1] = (byte)(iNum & 0x000000FF);
        }

        public static void ByteNumWrite(byte reg_add, ref byte[] reg_data, byte byte_count)
        // This method takes an array of bytes and writes registers numbered by byte_count
        // 3 byte format 0-exp, 1-msb, 2-lsb
        // 2 byte format 0-msb, 1-lsb
        {
            for (int a = 0; a < byte_count; ++a, ++reg_add)
                    Comm.I2C_write(reg_add, reg_data[a]);    // reg_data index 
        }
    }
}